import logging
import time
import typing
from dataclasses import dataclass, field

from starlette.applications import ASGIApp
from starlette.types import Receive, Scope, Send

Message = typing.MutableMapping[str, typing.Any]


@dataclass
class ResponseInterceptor:
    outer_send: Send
    response_message: Message | None = None

    async def send(self, message: Message) -> None:
        if message["type"] == "http.response.start":
            self.response_message = message
        await self.outer_send(message)

    @property
    def status_code(self) -> int:
        if self.response_message is None:
            return 500
        return self.response_message.get("status", 500)

    @property
    def content_length(self) -> int | None:
        res = self.get_header_value("content-length")
        if res is None:
            return None
        return int(res)

    @property
    def content_type(self) -> str | None:
        return self.get_header_value("content-type")

    def get_header_value(self, name: str) -> str | None:
        if self.response_message is None:
            return None
        for header in self.response_message.get("headers", []):
            print(header[0], name.encode("utf-8"))
            if header[0] == name.encode("utf-8"):
                try:
                    return header[1].decode("utf-8")
                except Exception:
                    pass
        return None


@dataclass
class ParsedScope:
    scope: Scope

    @property
    def method(self) -> str:
        method = self.scope.get("method")
        if method is None:
            raise ValueError("method is not set")
        return method

    @property
    def path(self) -> str:
        path = self.scope.get("path")
        if path is None:
            raise ValueError("path is not set")
        return path

    @property
    def query_string(self) -> str:
        query_string = self.scope.get("query_string")
        if query_string is None:
            return ""
        return query_string.decode("utf-8").strip()

    @property
    def full_path(self) -> str:
        qs = self.query_string
        if qs:
            return f"{self.path}?{qs}"
        return self.path


@dataclass
class AccessLogMiddleware:
    """
    Middleware to log access logs with a structured logger.

    Args:
        app: The ASGI application to wrap.
        logger: The structured logger to use.
    """

    app: ASGIApp
    logger: logging.LoggerAdapter
    ignore_hook: typing.Callable[[Scope], bool] = field(
        default_factory=lambda: lambda _: False
    )
    add_duration_s: bool = field(default=True)
    add_duration_ms: bool = field(default=False)
    add_duration_us: bool = field(default=False)
    add_response_content_length: bool = field(default=True)
    add_response_content_type: bool = field(default=True)
    add_response_headers: dict[str, str] = field(default_factory=dict)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            return await self.app(scope, receive, send)  # pragma: no cover
        if self.ignore_hook(scope):
            # ignored
            return await self.app(scope, receive, send)
        parsed_scope = ParsedScope(scope=scope)
        response_interceptor = ResponseInterceptor(outer_send=send)
        try:
            before = time.perf_counter()
            await self.app(scope, receive, response_interceptor.send)
        finally:
            after = time.perf_counter()
            duration = after - before
            if response_interceptor.status_code >= 400:
                log_fn = self.logger.warning
            else:
                log_fn = self.logger.info
            kwargs: dict[str, typing.Any] = {
                "method": parsed_scope.method,
                "full_path": parsed_scope.full_path,
                "status_code": response_interceptor.status_code,
            }
            if self.add_duration_s:
                kwargs["duration"] = duration
            if self.add_duration_ms:
                kwargs["duration_ms"] = duration * 1000.0
            if self.add_duration_us:
                kwargs["duration_us"] = int(duration * 1000000)
            if self.add_response_content_length:
                kwargs["response_content_length"] = response_interceptor.content_length
            if self.add_response_content_type:
                kwargs["response_content_type"] = response_interceptor.content_type
            if self.add_response_content_length:
                kwargs["response_content_length"] = response_interceptor.content_length
            if self.add_response_content_type:
                kwargs["response_content_type"] = response_interceptor.content_type
            for k, v in self.add_response_headers.items():
                value = response_interceptor.get_header_value(k)
                if value is not None:
                    kwargs[v] = value
            log_fn("access log", **kwargs)



import io
import json
from fastapi import FastAPI
from fastapi.testclient import TestClient
import stlog

from stlog_fastapi_middlewares.exception import CustomExceptionHandlingMiddleware


app = FastAPI()
logger = stlog.getLogger("test")
app.add_middleware(CustomExceptionHandlingMiddleware, logger=logger)
client = TestClient(app)


@app.get("/exception")
async def exception():
    raise Exception("test")


def test_exception(log_output: io.StringIO):
    response = client.get("/exception")
    assert response.status_code == 500
    print(log_output.getvalue())
    decoded = json.loads(log_output.getvalue())
    assert decoded["message"] == "an exception occurred"
    assert decoded["level"] == "ERROR"
    assert "Traceback" in decoded["exc_info"]
    assert decoded["source"]["lineno"] > 0



import logging
from dataclasses import dataclass

from starlette.applications import ASGIApp
from starlette.types import Receive, Scope, Send


@dataclass
class CustomExceptionHandlingMiddleware:
    """
    Middleware to log exceptions to structured logs.

    Args:
        app: The ASGI application to wrap.
        logger: The structured logger to use.
    """

    app: ASGIApp
    logger: logging.LoggerAdapter

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return None
        try:
            await self.app(scope, receive, send)
        except Exception:
            self.logger.error("an exception occurred", exc_info=True)
            await send(
                {
                    "type": "http.response.start",
                    "status": 500,
                    "headers": [(b"content-type", b"text/plain")],
                }
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": b"An unexpected error occurred. Please try again later.",
                    "more_body": False,
                }
            )



import io
import json
from fastapi import FastAPI, Response
from fastapi.testclient import TestClient
import stlog
from starlette.types import Scope

from stlog_fastapi_middlewares.access_log import AccessLogMiddleware


app = FastAPI()
logger = stlog.getLogger("test")
app.add_middleware(
    AccessLogMiddleware, logger=logger, add_response_headers={"x-test": "foo"}
)
client = TestClient(app)


def ignore_hook(scope: Scope) -> bool:
    print(scope.get("path"))
    return scope.get("path") == "/foo"


app_with_ignore_hook = FastAPI()
app_with_ignore_hook.add_middleware(
    AccessLogMiddleware,
    logger=logger,
    ignore_hook=ignore_hook,
)
client_with_ignore_hook = TestClient(app_with_ignore_hook)


@app.get("/foo")
async def foo(response: Response):
    response.headers["X-Test"] = "bar"
    return {"hello", "world"}


@app_with_ignore_hook.get("/foo")
async def foo2():
    return {"hello", "world"}


def test_access_log(log_output: io.StringIO):
    response = client.get("/foo")
    assert response.status_code == 200
    decoded = json.loads(log_output.getvalue())
    assert decoded["full_path"] == "/foo"
    assert decoded["status_code"] == 200
    assert decoded["method"] == "GET"
    assert decoded["level"] == "INFO"
    assert decoded["message"] == "access log"


def test_access_log_with_ignore_hook(log_output: io.StringIO):
    response = client_with_ignore_hook.get("/foo")
    assert response.status_code == 200
    assert len(log_output.getvalue()) == 0


def test_access_log_with_response_headers(log_output: io.StringIO):
    response = client.get("/foo", headers={"X-Test": "test"})
    assert response.status_code == 200
    decoded = json.loads(log_output.getvalue())
    assert decoded["foo"] == "bar"



import io
import json
import os
from fastapi import FastAPI
from fastapi.testclient import TestClient
import pytest
import stlog

from stlog_fastapi_middlewares.context import LogContextMiddleware


app = FastAPI()
logger = stlog.getLogger("test")
app.add_middleware(
    LogContextMiddleware,
    logger=logger,
    add_kvs={"foo": "bar"},
    add_pid=True,
    add_request_id=True,
    headers_to_kvs={"x-test-id": "test_id"},
    envs_to_kvs={"FOO2": "foo2"},
)
client = TestClient(app)


@app.get("/foo")
async def foo():
    logger.info("hello world")
    return {"hello", "world"}


@pytest.fixture
def modified_env():
    environ_copy = {}
    for key, value in os.environ.items():
        environ_copy[key] = value
    os.environ["FOO2"] = "bar2"
    yield
    for key, value in environ_copy.items():
        os.environ[key] = value


def test_context(log_output: io.StringIO, modified_env):
    response = client.get("/foo", headers={"x-test-id": "test-123"})
    assert response.status_code == 200
    decoded = json.loads(log_output.getvalue())
    print(json.dumps(decoded, indent=4))
    assert decoded["message"] == "hello world"
    assert decoded["foo"] == "bar"
    assert decoded["foo2"] == "bar2"
    assert decoded["pid"] > 0
    assert len(decoded["request_id"]) > 10
    assert decoded["test_id"] == "test-123"



from stlog_fastapi_middlewares.access_log import AccessLogMiddleware
from stlog_fastapi_middlewares.context import LogContextMiddleware
from stlog_fastapi_middlewares.exception import CustomExceptionHandlingMiddleware

__all__ = [
    "AccessLogMiddleware",
    "LogContextMiddleware",
    "CustomExceptionHandlingMiddleware",
]



from dataclasses import field, dataclass
import logging
import os
from typing import Any
from starlette.applications import ASGIApp
from starlette.types import Receive, Scope, Send
import stlog
import uuid


@dataclass()
class LogContextMiddleware:
    """
    Middleware to add request context to structured logs.

    Args:
        app: The ASGI application to wrap.
        logger: The logger to use.
        add_request_id: Whether to add a (unique) request ID to structured logs.
        add_pid: Whether to add a process ID to structured logs.
        add_kvs: Additional key-value pairs to add to structured logs.
        headers_to_kvs: Mapping of header names to key names in structured logs.
        envs_to_kvs: Mapping of environment variable names to key names in structured logs.
    """

    app: ASGIApp
    logger: logging.LoggerAdapter
    add_request_id: bool = True
    add_pid: bool = False
    add_kvs: dict[str, Any] = field(default_factory=dict)
    headers_to_kvs: dict[str, str] = field(default_factory=dict)
    envs_to_kvs: dict[str, str] = field(default_factory=dict)

    def __post_init__(self):
        for key, value in self.headers_to_kvs.items():
            self.headers_to_kvs[key.lower()] = value
        for key, value in self.envs_to_kvs.items():
            self.envs_to_kvs[key.upper()] = value

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return None
        kwargs: dict[str, Any] = {}
        if self.add_request_id:
            kwargs["request_id"] = str(uuid.uuid4()).replace("-", "")
        if self.add_pid:
            kwargs["pid"] = os.getpid()
        if self.add_kvs:
            kwargs.update(self.add_kvs)
        for k, v in self.envs_to_kvs.items():
            if k in os.environ:
                kwargs[v] = os.environ[k]
        headers = scope.get("headers", [])
        for name, value in headers:
            if name.decode("utf-8").lower() in self.headers_to_kvs:
                kwargs[self.headers_to_kvs[name.decode("utf-8").lower()]] = (
                    value.decode("utf-8")
                )
        with stlog.LogContext.bind(**kwargs):
            await self.app(scope, receive, send)
